from abc import ABCMeta, abstractmethod
from typing import Any

from jellyfish import (
    jaccard_similarity,
    jaro_similarity,
    jaro_winkler_similarity,
    levenshtein_distance,
)
from matchescu.matching.similarity._common import Similarity


class StringSimilarity(Similarity, metaclass=ABCMeta):
    def __init__(self, ignore_case: bool = False):
        self.__ignore_case = ignore_case

    @abstractmethod
    def _compute_string_similarity(self, x: str, m: int, y: str, n: int) -> float:
        pass

    def _compute_similarity(self, a: Any, b: Any) -> float:
        x = str(a)
        y = str(b)
        m = len(x)
        n = len(y)

        if m == 0 and n == 0:
            return 1
        if m == 0 or n == 0:
            return 0
        if self.__ignore_case:
            x = x.lower()
            y = y.lower()
        return self._compute_string_similarity(x, m, y, n)


class Levenshtein(StringSimilarity):
    def _compute_string_similarity(self, x: str, m: int, y: str, n: int) -> float:
        d = levenshtein_distance(x, y)
        res = 1 - d / max(m, n)
        return round(res, 2)


class Jaro(StringSimilarity):
    def _compute_string_similarity(self, x: str, m: int, y: str, n: int) -> float:
        return jaro_similarity(x, y)


class JaroWinkler(StringSimilarity):
    def _compute_string_similarity(self, x: str, m: int, y: str, n: int) -> float:
        return jaro_winkler_similarity(x, y)


class Jaccard(StringSimilarity):
    def __init__(self, ignore_case: bool = False, threshold: int | None = None):
        super().__init__(ignore_case)
        self.__threshold = threshold

    def _compute_string_similarity(self, x: str, m: int, y: str, n: int) -> float:
        threshold = self.__threshold or min(m, n)
        return jaccard_similarity(x, y, threshold)



from functools import partial
from typing import Any

import polars as pl

from matchescu.data import EntityReferenceExtraction
from matchescu.matching.entity_reference import (
    EntityReferenceComparisonConfig,
)
from matchescu.matching.ml.datasets._sampling import (
    AttributeComparison,
    PatternEncodedComparison,
)
from matchescu.typing import DataSource, Record


class DeduplicationDataSet:
    __TARGET_COL = "y"

    def __init__(
        self,
        source: DataSource[Record],
        ground_truth: set[tuple[Any, Any]],
    ) -> None:
        self.__extract = EntityReferenceExtraction(source, lambda ref: ref[0])
        self.__true_matches = ground_truth
        self.__comparison_data = None
        self.__sample_factory = None

    @property
    def target_vector(self) -> pl.DataFrame:
        if self.__comparison_data is None:
            raise ValueError("comparison matrix was not computed")
        return self.__comparison_data[self.__TARGET_COL]

    @property
    def feature_matrix(self) -> pl.DataFrame:
        if self.__comparison_data is None:
            raise ValueError("comparison matrix was not computed")
        return self.__comparison_data.drop([self.__TARGET_COL])

    @staticmethod
    def __with_col_suffix(
        extract: EntityReferenceExtraction, suffix: str
    ) -> pl.DataFrame:
        df = pl.DataFrame(extract())
        return df.rename({key: f"{key}{suffix}" for key in df.columns})

    def attr_compare(
        self, config: EntityReferenceComparisonConfig
    ) -> "DeduplicationDataSet":
        self.__sample_factory = AttributeComparison(
            self.__true_matches,
            config,
            self.__extract.identify,
            self.__extract.identify,
            self.__TARGET_COL,
        )
        return self

    def pattern_encoded(
        self, config: EntityReferenceComparisonConfig, possible_outcomes: int = 2
    ) -> "DeduplicationDataSet":
        self.__sample_factory = PatternEncodedComparison(
            self.__true_matches,
            config,
            self.__extract.identify,
            self.__extract.identify,
            self.__TARGET_COL,
            possible_outcomes,
        )
        return self

    def cross_sources(self) -> "DeduplicationDataSet":
        if self.__sample_factory is None:
            raise ValueError("specify type of sampling")
        source = list(self.__extract())
        if len(source) < 1:
            raise ValueError("no data")

        mid = len(source[0])
        data = []
        for i, left_row in enumerate(source):
            for j in range(i + 1, len(source)):
                row = {
                    f"left_column_{idx+1}": value for idx, value in enumerate(left_row)
                }
                row.update(
                    {
                        f"right_column_{idx + 1}": value
                        for idx, value in enumerate(source[j])
                    }
                )
                data.append(row)
        sample_factory = partial(self.__sample_factory, divider=mid)
        df = pl.DataFrame(data)
        self.__comparison_data = df.map_rows(sample_factory).unnest("column_0")
        return self



from decimal import Decimal
from numbers import Number
from typing import Any

from matchescu.matching.similarity import Similarity


class BoundedNumericDifferenceSimilarity(Similarity):
    __SUPPORTED_TYPES = (int, float, Decimal)

    def __init__(self, max_diff: float = 1.0) -> None:
        self.__max_diff = max_diff

    def _compute_similarity(self, a: Any, b: Any) -> float:
        if not isinstance(a, self.__SUPPORTED_TYPES) or not isinstance(
            b, self.__SUPPORTED_TYPES
        ):
            return 0
        diff = abs(a - b)
        if diff > self.__max_diff:
            diff = self.__max_diff
        return 1 - (diff / self.__max_diff)



from matchescu.matching.entity_reference._comparison import (
    EntityReferenceComparisonConfig,
    FellegiSunterComparison,
    NaiveBayesComparison,
    RawComparison,
)
from matchescu.matching.entity_reference._attr_spec import AttrComparisonSpec

__all__ = [
    "AttrComparisonSpec",
    "EntityReferenceComparisonConfig",
    "FellegiSunterComparison",
    "NaiveBayesComparison",
    "RawComparison",
]



from abc import abstractmethod, ABCMeta
from typing import Any


class Similarity(metaclass=ABCMeta):
    @abstractmethod
    def _compute_similarity(self, a: Any, b: Any) -> float:
        pass

    def __call__(self, a: Any, b: Any) -> float:
        if a is None and b is None:
            return 1
        elif a is None or b is None:
            return 0
        else:
            return self._compute_similarity(a, b)



from matchescu.matching.ml.datasets._csv import CsvDataSource
from matchescu.matching.ml.datasets._deduplication import DeduplicationDataSet
from matchescu.matching.ml.datasets._record_linkage import RecordLinkageDataSet
from matchescu.matching.ml.datasets._sampling import (
    AttributeComparison,
    PatternEncodedComparison,
)
from matchescu.matching.ml.datasets._traits import Traits


__all__ = [
    "CsvDataSource",
    "DeduplicationDataSet",
    "RecordLinkageDataSet",
    "Traits",
    "AttributeComparison",
    "PatternEncodedComparison",
]



from os import PathLike
from typing import Iterable, Iterator

import polars as pl

from matchescu.typing import Trait, Record


class CsvDataSource:
    def __init__(self, name: str, traits: Iterable[Trait], has_header: bool = True):
        self.name = name
        self.traits = traits
        self._df = None
        self._has_header = has_header

    def read_csv(self, path: str | PathLike) -> "CsvDataSource":
        self._df = pl.read_csv(path, ignore_errors=True, has_header=self._has_header)
        return self

    def __iter__(self) -> Iterator[Record]:
        if self._df is None:
            return iter([])
        return self._df.iter_rows()

    def __len__(self) -> int:
        if self._df is None:
            return 0
        return self._df.shape[0]



import math
from abc import ABCMeta, abstractmethod
from itertools import product
from typing import Hashable, Callable, Generator

from matchescu.matching.entity_reference import (
    EntityReferenceComparisonConfig,
    AttrComparisonSpec,
)
from matchescu.typing import EntityReference


class Sampling(metaclass=ABCMeta):
    def __init__(
        self,
        ground_truth: set[tuple[Hashable, Hashable]],
        cmp_config: EntityReferenceComparisonConfig,
        left_id: Callable[[EntityReference], Hashable],
        right_id: Callable[[EntityReference], Hashable],
        target_col_name: str,
    ):
        self._gt = ground_truth
        self._config = cmp_config
        self._left_id = left_id
        self._right_id = right_id
        self._target_col = target_col_name

    @abstractmethod
    def _process_cross_join_record(
        self, left: EntityReference, right: EntityReference
    ) -> dict:
        pass

    def __call__(self, cross_join_row: tuple, divider: int) -> tuple[dict]:
        left_side = cross_join_row[:divider]
        right_side = cross_join_row[divider:]
        result = self._process_cross_join_record(left_side, right_side)
        result[self._target_col] = int(
            (self._left_id(left_side), self._right_id(right_side)) in self._gt
        )
        return (result,)  # need to return a tuple


class AttributeComparison(Sampling):
    @staticmethod
    def __compare_attr_values(
        left_ref: EntityReference,
        right_ref: EntityReference,
        config: AttrComparisonSpec,
    ) -> int:
        a = left_ref[config.left_ref_key]
        b = right_ref[config.right_ref_key]
        return config.match_strategy(a, b)

    def _process_cross_join_record(
        self, left: EntityReference, right: EntityReference
    ) -> dict:
        return {
            spec.label: self.__compare_attr_values(left, right, spec)
            for spec in self._config.specs
        }


class PatternEncodedComparison(Sampling):
    _BASE = 2

    def __init__(
        self,
        ground_truth: set[tuple[Hashable, Hashable]],
        cmp_config: EntityReferenceComparisonConfig,
        left_id: Callable[[EntityReference], Hashable],
        right_id: Callable[[EntityReference], Hashable],
        target_col_name: str,
        possible_outcomes: int = 2,
    ):
        super().__init__(ground_truth, cmp_config, left_id, right_id, target_col_name)
        self._possible_outcomes = possible_outcomes

    def _generate_binary_patterns(self) -> Generator[tuple, None, None]:
        possible_outcomes = tuple(range(self._possible_outcomes))
        yield from product(possible_outcomes, repeat=len(self._config))

    def _process_cross_join_record(
        self, left: EntityReference, right: EntityReference
    ) -> dict:
        comparison_results = [
            spec.match_strategy(left[spec.left_ref_key], right[spec.right_ref_key])
            for spec in self._config.specs
        ]
        sample = {}
        for pattern in self._generate_binary_patterns():
            pattern_value = 0
            for idx, current in enumerate(zip(pattern, comparison_results)):
                expectation, actual = current
                coefficient = math.pow(self._BASE, idx)
                pattern_value += expectation * actual * coefficient
            sample["".join(map(str, pattern))] = pattern_value
        return sample



from matchescu.matching.attribute._match import (
    BinarySimilarityMatch,
    BinarySimilarityMatchOnThreshold,
    RawMatch,
    SimilarityMatch,
    TernarySimilarityMatch,
    TernarySimilarityMatchOnThreshold,
)
from matchescu.matching.attribute._match_result import (
    BinaryResult,
    TernaryResult,
)
from matchescu.matching.attribute._typing import AttrMatchCallable

__all__ = [
    "AttrMatchCallable",
    "BinaryResult",
    "BinarySimilarityMatch",
    "BinarySimilarityMatchOnThreshold",
    "RawMatch",
    "SimilarityMatch",
    "TernaryResult",
    "TernarySimilarityMatch",
    "TernarySimilarityMatchOnThreshold",
]



from matchescu.matching.similarity._common import Similarity
from matchescu.matching.similarity._exact_match import ExactMatch
from matchescu.matching.similarity._learned_levenshtein import LevenshteinLearner
from matchescu.matching.similarity._numeric import BoundedNumericDifferenceSimilarity
from matchescu.matching.similarity._string import (
    StringSimilarity,
    Jaccard,
    Jaro,
    JaroWinkler,
    Levenshtein,
)


__all__ = [
    "BoundedNumericDifferenceSimilarity",
    "Similarity",
    "ExactMatch",
    "Jaccard",
    "Jaro",
    "JaroWinkler",
    "Levenshtein",
    "LevenshteinLearner",
]



import math
from abc import ABCMeta, abstractmethod
from functools import partial
from itertools import product
from typing import Any, Callable, Hashable, Generator

import polars as pl

from matchescu.data import EntityReferenceExtraction
from matchescu.matching.entity_reference import (
    AttrComparisonSpec,
    EntityReferenceComparisonConfig,
)
from matchescu.typing import DataSource, Record, EntityReference


class Sampling(metaclass=ABCMeta):
    def __init__(
        self,
        ground_truth: set[tuple[Hashable, Hashable]],
        cmp_config: EntityReferenceComparisonConfig,
        left_id: Callable[[EntityReference], Hashable],
        right_id: Callable[[EntityReference], Hashable],
        target_col_name: str,
    ):
        self._gt = ground_truth
        self._config = cmp_config
        self._left_id = left_id
        self._right_id = right_id
        self._target_col = target_col_name

    @abstractmethod
    def _process_cross_join_record(
        self, left: EntityReference, right: EntityReference
    ) -> dict:
        pass

    def __call__(self, cross_join_row: tuple, divider: int) -> tuple[dict]:
        left_side = cross_join_row[:divider]
        right_side = cross_join_row[divider:]
        result = self._process_cross_join_record(left_side, right_side)
        result[self._target_col] = int(
            (self._left_id(left_side), self._right_id(right_side)) in self._gt
        )
        return (result,)  # need to return a tuple


class AttributeComparison(Sampling):
    @staticmethod
    def __compare_attr_values(
        left_ref: EntityReference,
        right_ref: EntityReference,
        config: AttrComparisonSpec,
    ) -> int:
        a = left_ref[config.left_ref_key]
        b = right_ref[config.right_ref_key]
        return config.match_strategy(a, b)

    def _process_cross_join_record(
        self, left: EntityReference, right: EntityReference
    ) -> dict:
        return {
            spec.label: self.__compare_attr_values(left, right, spec)
            for spec in self._config.specs
        }


class PatternEncodedComparison(Sampling):
    _BASE = 2

    def __init__(
        self,
        ground_truth: set[tuple[Hashable, Hashable]],
        cmp_config: EntityReferenceComparisonConfig,
        left_id: Callable[[EntityReference], Hashable],
        right_id: Callable[[EntityReference], Hashable],
        target_col_name: str,
        possible_outcomes: int = 2,
    ):
        super().__init__(ground_truth, cmp_config, left_id, right_id, target_col_name)
        self._possible_outcomes = possible_outcomes

    def _generate_binary_patterns(self) -> Generator[tuple, None, None]:
        possible_outcomes = tuple(range(self._possible_outcomes))
        yield from product(possible_outcomes, repeat=len(self._config))

    def _process_cross_join_record(
        self, left: EntityReference, right: EntityReference
    ) -> dict:
        comparison_results = [
            spec.match_strategy(left[spec.left_ref_key], right[spec.right_ref_key])
            for spec in self._config.specs
        ]
        sample = {}
        for pattern in self._generate_binary_patterns():
            pattern_value = 0
            for idx, current in enumerate(zip(pattern, comparison_results)):
                expectation, actual = current
                coefficient = math.pow(self._BASE, idx)
                pattern_value += expectation * actual * coefficient
            sample["".join(map(str, pattern))] = pattern_value
        return sample


class RecordLinkageDataSet:
    __TARGET_COL = "y"

    def __init__(
        self,
        left: DataSource[Record],
        right: DataSource[Record],
        ground_truth: set[tuple[Any, Any]],
    ) -> None:
        self.__extract_left = EntityReferenceExtraction(left, lambda ref: ref[0])
        self.__extract_right = EntityReferenceExtraction(right, lambda ref: ref[0])
        self.__true_matches = ground_truth
        self.__comparison_data = None
        self.__sample_factory = None

    @property
    def target_vector(self) -> pl.DataFrame:
        if self.__comparison_data is None:
            raise ValueError("comparison matrix was not computed")
        return self.__comparison_data[self.__TARGET_COL]

    @property
    def feature_matrix(self) -> pl.DataFrame:
        if self.__comparison_data is None:
            raise ValueError("comparison matrix was not computed")
        return self.__comparison_data.drop([self.__TARGET_COL])

    @staticmethod
    def __with_col_suffix(
        extract: EntityReferenceExtraction, suffix: str
    ) -> pl.DataFrame:
        df = pl.DataFrame(extract())
        return df.rename({key: f"{key}{suffix}" for key in df.columns})

    def attr_compare(
        self, config: EntityReferenceComparisonConfig
    ) -> "RecordLinkageDataSet":
        self.__sample_factory = AttributeComparison(
            self.__true_matches,
            config,
            self.__extract_left.identify,
            self.__extract_right.identify,
            self.__TARGET_COL,
        )
        return self

    def pattern_encoded(
        self, config: EntityReferenceComparisonConfig, possible_outcomes: int = 2
    ) -> "RecordLinkageDataSet":
        self.__sample_factory = PatternEncodedComparison(
            self.__true_matches,
            config,
            self.__extract_left.identify,
            self.__extract_right.identify,
            self.__TARGET_COL,
            possible_outcomes,
        )
        return self

    def cross_sources(self) -> "RecordLinkageDataSet":
        if self.__sample_factory is None:
            raise ValueError("specify type of sampling")
        left = self.__with_col_suffix(self.__extract_left, "_left")
        right = self.__with_col_suffix(self.__extract_right, "_right")
        cross_join = left.join(right, how="cross")
        sample_factory = partial(self.__sample_factory, divider=len(left.columns))
        self.__comparison_data = cross_join.map_rows(sample_factory).unnest("column_0")
        return self



import locale

from typing import Any, Callable, Iterator, Union

from matchescu.typing import Record, Trait


def _process_string(value: Any) -> str:
    return str(value)


def _process_int(value: Any) -> int:
    return int(value)


def _process_float(value: Any) -> float:
    return float(value)


def _process_currency(value: Any) -> float | None:
    if value is None:
        return None
    str_val = str(value).replace(",", "").strip()
    return locale.atof(str_val.lstrip("$"))


class RecordExtractionTrait:
    def __init__(self, mapping: Callable[[Any], Any], keys: list[int | str]) -> None:
        self.__mapping = mapping
        self.__keys = keys

    def __call__(self, item: Record) -> tuple:
        return tuple(self.__mapping(item[key]) for key in self.__keys)


class Traits:
    def __init__(self):
        self._traits = []

    def string(self, keys: list[Union[int, str]]) -> "Traits":
        self._traits.append(RecordExtractionTrait(_process_string, keys))
        return self

    def int(self, keys: list[Union[int, str]]) -> "Traits":
        self._traits.append(RecordExtractionTrait(_process_int, keys))
        return self

    def float(self, keys: list[Union[int, str]]) -> "Traits":
        self._traits.append(RecordExtractionTrait(_process_float, keys))
        return self

    def currency(self, keys: list[Union[int, str]]) -> "Traits":
        self._traits.append(RecordExtractionTrait(_process_currency, keys))
        return self

    def __iter__(self) -> Iterator[Trait]:
        return iter(self._traits)



from typing import Any

from matchescu.matching.similarity._common import Similarity


class ExactMatch(Similarity):
    def _compute_similarity(self, a: Any, b: Any) -> float:
        return 1 if a == b else 0
